@@ -43,6 +43,16 @@ class OpenSSLGenerator(ConfigGenerator):
         'RC2-CBC': '!RC2',
         'NULL': '!eNULL:!aNULL'
     }
+    cipher_notany_multimap = {
+        # CBC is documented as SSL_DES, SSL_3DES, SSL_RC2, SSL_IDEA,
+        #                      SSL_AES128, SSL_AES256,
+        #                      SSL_CAMELLIA128, SSL_CAMELLIA256, SSL_SEED
+        '-CBC': {'DES-CBC', '3DES-CBC', 'RC2-CBC', 'IDEA-CBC',
+                 'AES-128-CBC', 'AES-256-CBC',
+                 'CAMELLIA-128-CBC', 'CAMELLIA-256-CBC', 'SEED-CBC'},
+        '-AESCCM': {'AES-128-CCM', 'AES-256-CCM'},
+        '-AESGCM': {'AES-128-GCM', 'AES-256-GCM'},
+    }
 
     key_exchange_map = {
         'RSA': 'kRSA',
@@ -76,13 +86,44 @@ class OpenSSLGenerator(ConfigGenerator):
     }
 
     ciphersuite_map = {
-        'AES-256-GCM': 'TLS_AES_256_GCM_SHA384',
-        'AES-128-GCM': 'TLS_AES_128_GCM_SHA256',
-        'CHACHA20-POLY1305': 'TLS_CHACHA20_POLY1305_SHA256',
-        'AES-128-CCM': 'TLS_AES_128_CCM_SHA256',
-        'AES-128-CCM8': 'TLS_AES_128_CCM_8_SHA256',
-        'GOST28147-TC26Z-CNT': 'GOST2012-GOST8912-GOST8912',
-        'GOST28147-CPA-CNT': 'GOST2001-GOST89-GOST89'
+        'TLS_AES_256_GCM_SHA384': {
+            'cipher': {'AES-256-GCM'},
+            'hash': {'SHA2-384'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_AES_128_GCM_SHA256': {
+            'cipher': {'AES-128-GCM'},
+            'hash': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_CHACHA20_POLY1305_SHA256': {
+            'cipher': {'CHACHA20-POLY1305'},
+            'hash': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_AES_128_CCM_SHA256': {
+            'cipher': {'AES-128-CCM'},
+            'hash': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        # This one is not enableable since c-p does not expose CCM8 ciphers:
+        # 'TLS_AES_128_CCM_8_SHA256': {
+        #     'cipher': {'AES-128-CCM8'},  # this is not a thing in c-p
+        #     'hash': {'SHA2-256'},
+        #     'protocol': {'TLS1.3', 'DTLS1.3'},
+        # },
+        'TLS_SHA256_SHA256': {
+            'cipher': {'NULL'},
+            'hash': {'SHA2-256'},
+            'mac': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_SHA256_SHA384': {
+            'cipher': {'NULL'},
+            'hash': {'SHA2-384'},
+            'mac': {'SHA2-384'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
     }
 
     @classmethod
@@ -119,9 +160,9 @@ class OpenSSLGenerator(ConfigGenerator):
                 s = cls.append(s, cls.cipher_not_map[i])
             except KeyError:
                 pass
-        if 'AES-128-CCM' in ip['cipher']:
-            if 'AES-256-CCM' in ip['cipher']:
-                s = cls.append(s, '-AESCCM')
+        for keyword, cipherset in cls.cipher_notany_multimap.items():
+            if all(c in ip['cipher'] for c in cipherset):
+                s = cls.append(s, keyword)
 
         for i in ip['mac']:
             try:
@@ -141,12 +182,16 @@ class OpenSSLGenerator(ConfigGenerator):
     def generate_ciphersuites(cls, policy):
         s = ''
         p = policy.enabled
-        for i in p['cipher']:
-            try:
-                s = cls.append(s, cls.ciphersuite_map[i])
-            except KeyError:
-                pass
-
+        # we need the output sorted by order of `cipher`
+        # finer sorting nuances are currently ignored
+        for c in p['cipher']:
+            cipher_submap = {name: spec for name, spec
+                             in cls.ciphersuite_map.items()
+                             if spec['cipher'] == {c}}
+            for ciphersuite_name, ciphersuite_spec in cipher_submap.items():
+                if all(any(val in algvalues for val in p[algclass])
+                       for algclass, algvalues in ciphersuite_spec.items()):
+                    s = cls.append(s, ciphersuite_name)
         return s
 
     @classmethod
@@ -214,6 +259,10 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
         'RSA-PSS-RSAE-SHA2-512': 'rsa_pss_rsae_sha512',
         'EDDSA-ED25519': 'ed25519',
         'EDDSA-ED448': 'ed448',
+        # optional to not hard-require newer openssl version
+        'ECDSA-BRAINPOOLP256-SHA2-256': '?ecdsa_brainpoolP256r1_sha256',
+        'ECDSA-BRAINPOOLP384-SHA2-384': '?ecdsa_brainpoolP384r1_sha384',
+        'ECDSA-BRAINPOOLP512-SHA2-512': '?ecdsa_brainpoolP512r1_sha512',
         # provider-only, so, optional (openssl#23050) + marked experimental
         'MLDSA44': '?mldsa44',
         'P256-MLDSA44': '?p256_mldsa44',
@@ -270,9 +319,9 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
         'FFDHE-4096': 'ffdhe4096',
         'FFDHE-6144': 'ffdhe6144',
         'FFDHE-8192': 'ffdhe8192',
-        'BRAINPOOL-P256R1': 'brainpoolP256r1',
-        'BRAINPOOL-P384R1': 'brainpoolP384r1',
-        'BRAINPOOL-P512R1': 'brainpoolP512r1',
+        'BRAINPOOL-P256R1': 'brainpoolP256r1:?brainpoolP256r1tls13',
+        'BRAINPOOL-P384R1': 'brainpoolP384r1:?brainpoolP384r1tls13',
+        'BRAINPOOL-P512R1': 'brainpoolP512r1:?brainpoolP512r1tls13',
         # provider-only, so, optional (openssl#23050) + marked experimental
         'KYBER768': '?kyber768',
         'X25519-KYBER768': '?x25519_kyber768',
@@ -283,8 +332,8 @@ class OpenSSLConfigGenerator(OpenSSLGenerator):
         'MLKEM768': '?mlkem768',
         'P384-MLKEM768': '?p384_mlkem768',
         'X448-MLKEM768': '?x448_mlkem768',
-        'X25519-MLKEM768': '?x25519_mlkem768',
-        'P256-MLKEM768': '?p256_mlkem768',
+        'X25519-MLKEM768': '?X25519MLKEM768:?x25519_mlkem768',  # old/new name
+        'P256-MLKEM768': '?SecP256r1MLKEM768:?p256_mlkem768',  # old/new name
         'MLKEM1024': '?mlkem1024',
         'P521-MLKEM1024': '?p521_mlkem1024',
         'P384-MLKEM1024': '?p384_mlkem1024',
